from ._mpe_utils.simple_env import SimpleEnv, make_env
from .scenarios.simple_adversary import Scenario
from pettingzoo.utils.to_parallel import parallel_wrapper_fn


class raw_env(SimpleEnv):
    def __init__(self, seed=None, N=2, max_frames=100):
        scenario = Scenario()
        world = scenario.make_world(N=2)
        super().__init__(scenario, world, max_frames, seed)


env = make_env(raw_env)
parallel_env = parallel_wrapper_fn(env)



from pettingzoo.utils.error import DeprecatedEnv


def env(*args, **kwargs):
    raise DeprecatedEnv("prison_v0 is now depreciated, use prison_v1 instead")


raw_env = env
parallel_env = env
manual_control = env



from pettingzoo import AECEnv
from pettingzoo.utils.agent_selector import agent_selector
from gym import spaces
import rlcard
import random
from rlcard.games.uno.card import UnoCard
import numpy as np
from pettingzoo.utils import wrappers
from .rlcard_base import RLCardBase


def env(**kwargs):
    env = raw_env(**kwargs)
    env = wrappers.TerminateIllegalWrapper(env, illegal_reward=-1)
    env = wrappers.AssertOutOfBoundsWrapper(env)
    env = wrappers.NaNRandomWrapper(env)
    env = wrappers.OrderEnforcingWrapper(env)
    return env


class raw_env(RLCardBase):

    metadata = {'render.modes': ['human']}

    def __init__(self, seed=None):
        super().__init__("uno", 2, (7, 4, 15), seed)

    def render(self, mode='human'):
        for player in self.agents:
            state = self.env.game.get_state(self._name_to_int(player))
            print("\n\n=============== {}'s Hand ===============".format(player))
            UnoCard.print_cards(state['hand'])
        print('\n\n================= Target Card =================')
        UnoCard.print_cards(state['target'], wild_color=True)
        print('\n')

    def close(self):
        pass



from .waterworld_base import MAWaterWorld as _env
from pettingzoo import AECEnv
from pettingzoo.utils import agent_selector
from pettingzoo.utils import wrappers
import numpy as np
from pettingzoo.utils.to_parallel import parallel_wrapper_fn


def env(**kwargs):
    env = raw_env(**kwargs)
    env = wrappers.ClipOutOfBoundsWrapper(env)
    env = wrappers.NanZerosWrapper(env)
    env = wrappers.OrderEnforcingWrapper(env)
    return env


parallel_env = parallel_wrapper_fn(env)


class raw_env(AECEnv):

    metadata = {'render.modes': ['human']}

    def __init__(self, seed=None, *args, **kwargs):
        super().__init__()
        self.env = _env(seed, *args, **kwargs)

        self.num_agents = self.env.num_agents
        self.agents = ["pursuer_" + str(r) for r in range(self.num_agents)]
        self.agent_name_mapping = dict(zip(self.agents, list(range(self.num_agents))))
        self._agent_selector = agent_selector(self.agents)
        # spaces
        self.action_spaces = dict(zip(self.agents, self.env.action_space))
        self.observation_spaces = dict(
            zip(self.agents, self.env.observation_space))
        self.steps = 0
        self.display_wait = 0.03
        self.has_reset = False

    def convert_to_dict(self, list_of_list):
        return dict(zip(self.agents, list_of_list))

    def reset(self, observe=True):
        self.has_reset = True
        self.steps = 0
        self.env.reset()
        self._agent_selector.reinit(self.agents)
        self.agent_selection = self._agent_selector.next()
        self.rewards = dict(
            zip(self.agents, [np.float64(0) for _ in self.agents]))
        self.dones = dict(zip(self.agents, [False for _ in self.agents]))
        self.infos = dict(zip(self.agents, [{} for _ in self.agents]))
        if observe:
            return self.observe(self.agent_selection)

    def close(self):
        if self.has_reset:
            self.env.close()

    def render(self, mode="human"):
        self.env.render()

    def step(self, action, observe=True):
        agent = self.agent_selection

        self.env.step(action, self.agent_name_mapping[agent], self._agent_selector.is_last())
        for r in self.rewards:
            self.rewards[r] = self.env.last_rewards[self.agent_name_mapping[r]]

        if self.env.frames >= self.env.max_frames:
            self.dones = dict(zip(self.agents, [True for _ in self.agents]))
        else:
            self.dones = dict(zip(self.agents, self.env.last_dones))
        self.agent_selection = self._agent_selector.next()

        # AGENT SELECT

        self.steps += 1

        if observe:
            return self.observe(self.agent_selection)

    def observe(self, agent):
        return self.env.observe(self.agent_name_mapping[agent])



from pettingzoo.utils import agent_selector
from pettingzoo import AECEnv


class base_parallel_env:
    def __init__(self):
        '''
        sets the following attributes:

        agents: list of agent names
        num_agents: number of agent names
        action_spaces: list of action spaces corresponding to each agent
        observation_spaces: list of observation spaces corresponding to each agent
        '''

    def render(self):
        '''
        render the environment.
        '''
        raise NotImplementedError()

    def close(self):
        '''
        release rendering or other resources
        '''
        raise NotImplementedError()

    def reset(self):
        '''
        resets environment
        return observations for all agents
        '''
        raise NotImplementedError()

    def step(self, all_actions):
        '''
        step using list of actions for all agents,
        including those that are already done (those actions are ignored)

        return observations, rewards, dones, infos
        '''
        raise NotImplementedError()


class _parallel_env_wrapper(AECEnv):

    metadata = {'render.modes': ['human']}

    def __init__(self, parallel_env):
        self.env = parallel_env
        self.agents = self.env.agents
        self.num_agents = len(self.agents)

        self.action_spaces = {agent: space for agent, space in zip(self.agents, self.env.action_spaces)}
        self.observation_spaces = {agent: space for agent, space in zip(self.agents, self.env.observation_spaces)}
        self._agent_mapper = {agent: i for i, agent in enumerate(self.agents)}

    def reset(self, observe=True):
        self._actions = [None] * self.num_agents

        self._live_agents = self.agents[:]
        self._agent_selector = agent_selector(self._live_agents)
        self.agent_selection = self._agent_selector.reset()
        self.dones = {agent: False for agent in self.agents}
        self.infos = {agent: {} for agent in self.agents}
        self.rewards = {agent: 0 for agent in self.agents}

        self._observations = self.env.reset()

        return self.observe(self.agent_selection) if observe else None

    def observe(self, agent):
        return self._observations[self._agent_mapper[agent]]

    def step(self, action, observe=True):
        self._actions[self._agent_mapper[self.agent_selection]] = action
        if self._agent_selector.is_last():
            obss, rews, dones, infos = self.env.step(self._actions)
            self._observations = obss

            self.dones = {agent: done for agent, done in zip(self.agents, dones)}
            self.infos = {agent: info for agent, info in zip(self.agents, infos)}
            self.rewards = {agent: reward for agent, reward in zip(self.agents, rews)}

            self._live_agents = [agent for done, agent in zip(dones, self.agents) if not done]
            if len(self._live_agents):
                self._agent_selector = agent_selector(self._live_agents)
                self.agent_selection = self._agent_selector.reset()
        else:
            self.agent_selection = self._agent_selector.next()
        return self.observe(self.agent_selection) if observe else None

    def render(self, mode="human"):
        self.env.render()

    def close(self):
        self.env.close()



from pettingzoo.utils.env import AECEnv
from pettingzoo.utils._parallel_env import _parallel_env_wrapper
from pettingzoo.utils.wrappers import OrderEnforcingWrapper


class ParallelEnv:
    def reset(self):
        '''
        returns:
        observations a dictionary {"agent_1":obs_1, ..., "agent_n": obs_n}
        '''

    def step(self, actions):
        '''
        parameters:
        - actions: {"agent_1": action1, ..., "agent_n": action_n}

        returns:
        (observations, rewards, dones, infos)

        steps all active agents in the environment, returns all observations, rewards, infos of those active agentts
        which are all dictionaries
        '''

    def render(self, mode="human"):
        '''
        renders environment by printing text or through a window
        '''

    def close(self):
        '''
        releases resources (typically just closes the rendering window)
        '''


class to_parallel(ParallelEnv):
    def __init__(self, aec_env):
        self.aec_env = aec_env
        self.observation_spaces = aec_env.observation_spaces
        self.action_spaces = aec_env.action_spaces
        self.agents = aec_env.agents
        self.num_agents = aec_env.num_agents
        self._was_dones = {agent: False for agent in self.agents}

    def reset(self):
        self.aec_env.reset(observe=False)
        self._was_dones = {agent: False for agent in self.agents}
        observations = {agent: self.aec_env.observe(agent) for agent in self.aec_env.agents if not self.aec_env.dones[agent]}
        return observations

    def step(self, actions):
        rewards = {}
        dones = {}
        infos = {}

        for agent in self.agents:
            if not self._was_dones[agent]:
                assert agent == self.aec_env.agent_selection, f"environment has a nontrivial ordering, and cannot be used with the to_parallel wrapper\nCurrent agent: {self.aec_env.agent_selection}\nExpected agent: {agent}"
                assert agent in actions, "Live environment agent is not in actions dictionary"
                self._was_dones[agent] = self.aec_env.dones[agent]
                self.aec_env.step(actions[agent], observe=False)
                agent = self.aec_env.agent_selection

        rewards = self.aec_env.rewards
        dones = self.aec_env.dones
        infos = self.aec_env.infos
        observations = {agent: self.aec_env.observe(agent) for agent in self.aec_env.agents}
        return observations, rewards, dones, infos

    def render(self, mode="human"):
        return self.aec_env.render(mode)

    def close(self):
        return self.aec_env.close()


def parallel_wrapper_fn(env_fn):
    def par_fn(**kwargs):
        env = env_fn(**kwargs)
        env = to_parallel(env)
        return env
    return par_fn


class Sequentialize:
    def __init__(self, par_env):
        self.agents = par_env.agents
        self.observation_spaces = [par_env.observation_spaces[agent] for agent in self.agents]
        self.action_spaces = [par_env.action_spaces[agent] for agent in self.agents]
        self.par_env = par_env

    def _sequentialize(self, d):
        return [d.get(agent, None) for agent in self.agents]

    def reset(self):
        obs_dict = self.par_env.reset()

        return self._sequentialize(obs_dict)

    def step(self, actions):
        act_list = {agent: actions[i] for i, agent in enumerate(self.agents) if actions[i] is not None}
        obs, rew, done, info = self.par_env.step(act_list)
        obs = self._sequentialize(obs)
        rew = self._sequentialize(rew)
        done = self._sequentialize(done)
        info = self._sequentialize(info)
        return obs, rew, done, info

    def render(self, mode="human"):
        return self.par_env.render(mode)

    def close(self):
        self.par_env.close()


def from_parallel(par_env):
    if isinstance(par_env, to_parallel):
        return par_env.aec_env
    else:
        sequ_env = Sequentialize(par_env)
        aec_env = _parallel_env_wrapper(sequ_env)
        ordered_env = OrderEnforcingWrapper(aec_env)
        return ordered_env
